"""
winerp
~~~~~~

An extension module to facilitate creation of bot commands.

:copyright: (c) 2022-present BlackThunder
:license: MIT, see LICENSE for more details.
"""

from .client import Client
from .server import Server
from .lib.payload import winerpObject
